@@ -888,8 +888,8 @@ report_discard_pending(int pending, void *value)
 }
 
 /*
- * Return TRUE if "arg" is only a variable, register, environment variable or
- * option name.
+ * Return TRUE if "arg" is only a variable, register, environment variable,
+ * option name or string.
  */
     int
 cmd_is_name_only(char_u *arg)
@@ -904,6 +904,17 @@ cmd_is_name_only(char_u *arg)
 	if (*p != NUL)
 	    ++p;
     }
+    else if (*p == '\'' || *p == '"')
+    {
+	int	    r;
+
+	if (*p == '"')
+	    r = eval_string(&p, NULL, FALSE, FALSE);
+	else
+	    r = eval_lit_string(&p, NULL, FALSE, FALSE);
+	if (r == FAIL)
+	    return FALSE;
+    }
     else
     {
 	if (*p == '&')
@@ -940,8 +951,14 @@ ex_eval(exarg_T *eap)
     if (eval0(eap->arg, &tv, eap, &evalarg) == OK)
     {
 	clear_tv(&tv);
-	if (in_vim9script() && name_only && lnum == SOURCING_LNUM)
+	if (in_vim9script() && name_only
+		&& (evalarg.eval_tofree == NULL
+		    || ends_excmd2(evalarg.eval_tofree,
+					      skipwhite(evalarg.eval_tofree))))
+	{
+	    SOURCING_LNUM = lnum;
 	    semsg(_(e_expression_without_effect_str), eap->arg);
+	}
     }
 
     clear_evalarg(&evalarg, eap);
@@ -1118,6 +1135,14 @@ ex_else(exarg_T *eap)
 	skip = TRUE;
     }
 
+    if (cstack->cs_idx >= 0)
+    {
+	// Variables declared in the previous block can no longer be
+	// used.  Needs to be done before setting "cs_flags".
+	leave_block(cstack);
+	enter_block(cstack);
+    }
+
     // if skipping or the ":if" was TRUE, reset ACTIVE, otherwise set it
     if (skip || cstack->cs_flags[cstack->cs_idx] & CSF_TRUE)
     {
@@ -1146,7 +1171,12 @@ ex_else(exarg_T *eap)
 
     if (eap->cmdidx == CMD_elseif)
     {
-	result = eval_to_bool(eap->arg, &error, eap, skip);
+	// When skipping we ignore most errors, but a missing expression is
+	// wrong, perhaps it should have been "else".
+	if (skip && ends_excmd(*eap->arg))
+	    semsg(_(e_invalid_expression_str), eap->arg);
+	else
+	    result = eval_to_bool(eap->arg, &error, eap, skip);
 
 	// When throwing error exceptions, we want to throw always the first
 	// of several errors in a row.  This is what actually happens when
