@@ -293,6 +293,7 @@ prt_get_attr(
     pattr->italic = (highlight_has_attr(hl_id, HL_ITALIC, modec) != NULL);
     pattr->underline = (highlight_has_attr(hl_id, HL_UNDERLINE, modec) != NULL);
     pattr->undercurl = (highlight_has_attr(hl_id, HL_UNDERCURL, modec) != NULL);
+    // TODO: HL_UNDERDOUBLE, HL_UNDERDOTTED, HL_UNDERDASHED
 
 # if defined(FEAT_GUI) || defined(FEAT_TERMGUICOLORS)
     if (USE_24BIT)
@@ -1419,9 +1420,6 @@ prt_write_file(char_u *buffer)
     static void
 prt_write_file_len(char_u *buffer, int bytes)
 {
-#ifdef EBCDIC
-    ebcdic2ascii(buffer, bytes);
-#endif
     prt_write_file_raw_len(buffer, bytes);
 }
 
@@ -1626,8 +1624,6 @@ prt_flush_buffer(void)
 	    prt_write_string("ul\n");
 	}
 	// Draw the text
-	// Note: we write text out raw - EBCDIC conversion is handled in the
-	// PostScript world via the font encoding vector.
 	if (prt_out_mbyte)
 	    prt_write_string("<");
 	else
@@ -1650,7 +1646,7 @@ prt_flush_buffer(void)
 	    prt_write_string("s\n");
 
 	ga_clear(&prt_ps_buffer);
-	ga_init2(&prt_ps_buffer, (int)sizeof(char), prt_bufsiz);
+	ga_init2(&prt_ps_buffer, sizeof(char), prt_bufsiz);
     }
 }
 
@@ -1839,7 +1835,7 @@ prt_open_resource(struct prt_ps_resource_S *resource)
     fd_resource = mch_fopen((char *)resource->filename, READBIN);
     if (fd_resource == NULL)
     {
-	semsg(_("E624: Can't open file \"%s\""), resource->filename);
+	semsg(_(e_cant_open_file_str_3), resource->filename);
 	return FALSE;
     }
     CLEAR_FIELD(prt_resfile.buffer);
@@ -1866,8 +1862,8 @@ prt_open_resource(struct prt_ps_resource_S *resource)
     if (prt_resfile_strncmp(offset, PRT_RESOURCE_HEADER,
 				       (int)STRLEN(PRT_RESOURCE_HEADER)) != 0)
     {
-	semsg(_("E618: file \"%s\" is not a PostScript resource file"),
-		resource->filename);
+	semsg(_(e_file_str_is_not_postscript_resource_file),
+							   resource->filename);
 	return FALSE;
     }
 
@@ -1883,8 +1879,8 @@ prt_open_resource(struct prt_ps_resource_S *resource)
     if (prt_resfile_strncmp(offset, PRT_RESOURCE_RESOURCE,
 				     (int)STRLEN(PRT_RESOURCE_RESOURCE)) != 0)
     {
-	semsg(_("E619: file \"%s\" is not a supported PostScript resource file"),
-		resource->filename);
+	semsg(_(e_file_str_is_not_supported_postscript_resource_file),
+							   resource->filename);
 	return FALSE;
     }
     offset += (int)STRLEN(PRT_RESOURCE_RESOURCE);
@@ -1901,8 +1897,8 @@ prt_open_resource(struct prt_ps_resource_S *resource)
 	resource->type = PRT_RESOURCE_TYPE_CMAP;
     else
     {
-	semsg(_("E619: file \"%s\" is not a supported PostScript resource file"),
-		resource->filename);
+	semsg(_(e_file_str_is_not_supported_postscript_resource_file),
+							   resource->filename);
 	return FALSE;
     }
 
@@ -1943,8 +1939,8 @@ prt_open_resource(struct prt_ps_resource_S *resource)
 
     if (!seen_title || !seen_version)
     {
-	semsg(_("E619: file \"%s\" is not a supported PostScript resource file"),
-		resource->filename);
+	semsg(_(e_file_str_is_not_supported_postscript_resource_file),
+							   resource->filename);
 	return FALSE;
     }
 
@@ -1957,7 +1953,7 @@ prt_check_resource(struct prt_ps_resource_S *resource, char_u *version)
     // Version number m.n should match, the revision number does not matter
     if (STRNCMP(resource->version, version, STRLEN(version)))
     {
-	semsg(_("E621: \"%s\" resource file has wrong version"),
+	semsg(_(e_str_resource_file_has_wrong_version),
 		resource->name);
 	return FALSE;
     }
@@ -2417,7 +2413,7 @@ mch_print_init(
 	    // Check encoding and character set are compatible
 	    if ((p_mbenc->needs_charset & p_mbchar->has_charset) == 0)
 	    {
-		emsg(_("E673: Incompatible multi-byte encoding and character set."));
+		emsg(_(e_incompatible_multi_byte_encoding_and_character_set));
 		return FALSE;
 	    }
 
@@ -2434,7 +2430,7 @@ mch_print_init(
 	    // Add custom CMap character set name
 	    if (*p_pmcs == NUL)
 	    {
-		emsg(_("E674: printmbcharset cannot be empty with multi-byte encoding."));
+		emsg(_(e_printmbcharset_cannot_be_empty_with_multi_byte_encoding));
 		return FALSE;
 	    }
 	    vim_strncpy((char_u *)prt_cmap, p_pmcs, sizeof(prt_cmap) - 3);
@@ -2452,7 +2448,7 @@ mch_print_init(
 
 	if (!mbfont_opts[OPT_MBFONT_REGULAR].present)
 	{
-	    emsg(_("E675: No default font specified for multi-byte printing."));
+	    emsg(_(e_no_default_font_specified_for_multi_byte_printing));
 	    return FALSE;
 	}
 
@@ -2649,7 +2645,7 @@ mch_print_init(
     prt_bufsiz = psettings->chars_per_line;
     if (prt_out_mbyte)
 	prt_bufsiz *= 2;
-    ga_init2(&prt_ps_buffer, (int)sizeof(char), prt_bufsiz);
+    ga_init2(&prt_ps_buffer, sizeof(char), prt_bufsiz);
 
     prt_page_num = 0;
 
@@ -2769,9 +2765,9 @@ mch_print_begin(prt_settings_T *psettings)
 	// derive the bbox from that point.  We have the expected cpl chars
 	// across the media and lpp lines down the media.
 	bbox[1] = (int)(top - (psettings->lines_per_page + prt_header_height())
-							    * prt_line_height);
-	bbox[2] = (int)(left + psettings->chars_per_line * prt_char_width
-									+ 0.5);
+						    * (double)prt_line_height);
+	bbox[2] = (int)(left + psettings->chars_per_line
+					       * (double)prt_char_width + 0.5);
 	bbox[3] = (int)(top + 0.5);
     }
     else
@@ -2782,8 +2778,8 @@ mch_print_begin(prt_settings_T *psettings)
 	bbox[1] = (int)bottom;
 	bbox[2] = (int)(left + ((psettings->lines_per_page
 			      + prt_header_height()) * prt_line_height) + 0.5);
-	bbox[3] = (int)(bottom + psettings->chars_per_line * prt_char_width
-									+ 0.5);
+	bbox[3] = (int)(bottom + psettings->chars_per_line
+					       * (double)prt_char_width + 0.5);
     }
     prt_dsc_ints("BoundingBox", 4, bbox);
     // The media width and height does not change with landscape printing!
@@ -2797,7 +2793,7 @@ mch_print_begin(prt_settings_T *psettings)
     if (prt_out_mbyte)
     {
 	prt_dsc_font_resource((prt_use_courier ? NULL
-				 : "DocumentNeededResources"), &prt_ps_mb_font);
+				: "DocumentNeededResources"), &prt_ps_mb_font);
 	if (!prt_custom_cmap)
 	    prt_dsc_resources(NULL, "cmap", prt_cmap);
     }
@@ -2882,12 +2878,12 @@ mch_print_begin(prt_settings_T *psettings)
     }
 
     prt_conv.vc_type = CONV_NONE;
-    if (!(enc_canon_props(p_enc) & enc_canon_props(p_encoding) & ENC_8BIT)) {
+    if (!(enc_canon_props(p_enc) & enc_canon_props(p_encoding) & ENC_8BIT))
+    {
 	// Set up encoding conversion if required
 	if (FAIL == convert_setup(&prt_conv, p_enc, p_encoding))
 	{
-	    semsg(_("E620: Unable to convert to print encoding \"%s\""),
-		    p_encoding);
+	    semsg(_(e_unable_to_convert_to_print_encoding_str), p_encoding);
 	    goto theend;
 	}
 	prt_do_conv = TRUE;
@@ -3119,7 +3115,7 @@ mch_print_end(prt_settings_T *psettings)
 
     // Write CTRL-D to close serial communication link if used.
     // NOTHING MUST BE WRITTEN AFTER THIS!
-    prt_write_file((char_u *)IF_EB("\004", "\067"));
+    prt_write_file((char_u *)"\004");
 
     if (!prt_file_error && psettings->outfile == NULL
 					&& !got_int && !psettings->user_abort)
@@ -3379,26 +3375,21 @@ mch_print_text_out(char_u *textp, int len UNUSED)
 	{
 	    // Convert non-printing characters to either their escape or octal
 	    // sequence, ensures PS sent over a serial line does not interfere
-	    // with the comms protocol.  Note: For EBCDIC we need to write out
-	    // the escape sequences as ASCII codes!
-	    // Note 2: Char codes < 32 are identical in EBCDIC and ASCII AFAIK!
-	    ga_append(&prt_ps_buffer, IF_EB('\\', 0134));
+	    // with the comms protocol.
+	    ga_append(&prt_ps_buffer, '\\');
 	    switch (ch)
 	    {
-		case BS:   ga_append(&prt_ps_buffer, IF_EB('b', 0142)); break;
-		case TAB:  ga_append(&prt_ps_buffer, IF_EB('t', 0164)); break;
-		case NL:   ga_append(&prt_ps_buffer, IF_EB('n', 0156)); break;
-		case FF:   ga_append(&prt_ps_buffer, IF_EB('f', 0146)); break;
-		case CAR:  ga_append(&prt_ps_buffer, IF_EB('r', 0162)); break;
-		case '(':  ga_append(&prt_ps_buffer, IF_EB('(', 0050)); break;
-		case ')':  ga_append(&prt_ps_buffer, IF_EB(')', 0051)); break;
-		case '\\': ga_append(&prt_ps_buffer, IF_EB('\\', 0134)); break;
+		case BS:   ga_append(&prt_ps_buffer, 'b'); break;
+		case TAB:  ga_append(&prt_ps_buffer, 't'); break;
+		case NL:   ga_append(&prt_ps_buffer, 'n'); break;
+		case FF:   ga_append(&prt_ps_buffer, 'f'); break;
+		case CAR:  ga_append(&prt_ps_buffer, 'r'); break;
+		case '(':  ga_append(&prt_ps_buffer, '('); break;
+		case ')':  ga_append(&prt_ps_buffer, ')'); break;
+		case '\\': ga_append(&prt_ps_buffer, '\\'); break;
 
 		default:
 			   sprintf((char *)ch_buff, "%03o", (unsigned int)ch);
-#ifdef EBCDIC
-			   ebcdic2ascii(ch_buff, 3);
-#endif
 			   ga_append(&prt_ps_buffer, ch_buff[0]);
 			   ga_append(&prt_ps_buffer, ch_buff[1]);
 			   ga_append(&prt_ps_buffer, ch_buff[2]);
